from typing import Generator

import requests
from iamcore.irn import IRN
from requests import Response

from iamcore.client.conf import IAMCORE_URL
from .common import SortOrder, to_snake_case, to_dict, generic_search_all, IamEntityResponse, IamEntitiesResponse
from .exceptions import IAMTenantException, err_chain, unwrap_post, unwrap_put, unwrap_delete, \
    unwrap_get, IAMException


class Tenant(object):
    resource_id: str
    irn: IRN
    tenant_id: str
    name: str
    display_name: str
    created: str
    updated: str

    def __init__(self, irn: str, **kwargs):
        self.irn = IRN.from_irn_str(irn)
        for k, v in kwargs.items():
            attr = to_snake_case(k)
            setattr(self, attr, v)

    @staticmethod
    def of(item):
        if isinstance(item, Tenant):
            return item
        elif isinstance(item, dict):
            return Tenant(**item)
        raise IAMTenantException(f"Unexpected response format")

    def to_dict(self):
        return to_dict(self)

    def update(self, auth_headers: dict[str, str]) -> None:
        return update_tenant(auth_headers, self.resource_id, self.display_name)

    def delete(self, auth_headers: dict[str, str]):
        return delete_tenant(auth_headers, self.resource_id)


class TenantIssuer(object):
    id: str
    irn: IRN
    name: str
    type: str
    url: str
    client_id: str
    login_url: str

    def __init__(self, irn: str, **kwargs):
        self.irn = IRN.from_irn_str(irn)
        for k, v in kwargs.items():
            attr = to_snake_case(k)
            setattr(self, attr, v)

    @staticmethod
    def of(item):
        if isinstance(item, TenantIssuer):
            return item
        elif isinstance(item, dict):
            return TenantIssuer(**item)
        raise IAMTenantException(f"Unexpected response format")

    def to_dict(self):
        return to_dict(self)

    def update(self, auth_headers: dict[str, str]) -> None:
        return update_tenant(auth_headers, self.resource_id, self.display_name)

    def delete(self, auth_headers: dict[str, str]):
        return delete_tenant(auth_headers, self.resource_id)


@err_chain(IAMTenantException)
def create_tenant(auth_headers: dict[str, str], payload: dict[str, str] = None,
                  name: str = None, display_name: str = None) -> Tenant:
    url = IAMCORE_URL + "/api/v1/tenants/issuer-types/iamcore"
    if not payload:
        payload = {
            "name": name,
            "displayName": display_name
        }
    headers = {
        "Content-Type": "application/json",
        **auth_headers
    }
    response: Response = requests.request("POST", url, json=payload, headers=headers)
    return IamEntityResponse(Tenant, **unwrap_post(response)).data


@err_chain(IAMTenantException)
def update_tenant(auth_headers: dict[str, str], resource_id: str, display_name: str) -> None:
    if not auth_headers:
        raise IAMTenantException(f"Missing authorization headers")
    if not resource_id or not display_name:
        raise IAMTenantException(f"Missing resource_id or display_name")

    url = IAMCORE_URL + "/api/v1/tenants/" + IRN.of(resource_id).to_base64()
    payload = {
        "displayName": display_name
    }
    headers = {
        "Content-Type": "application/json",
        **auth_headers
    }
    response: Response = requests.request("PUT", url, json=payload, headers=headers)
    return unwrap_put(response)


@err_chain(IAMTenantException)
def delete_tenant(auth_headers: dict[str, str], resource_id: str) -> None:
    if not auth_headers:
        raise IAMTenantException(f"Missing authorization headers")
    if not resource_id:
        raise IAMTenantException(f"Missing resource_id")

    url = IAMCORE_URL + "/api/v1/tenants/" + IRN.of(resource_id).to_base64()
    headers = {
        "Content-Type": "application/json",
        **auth_headers
    }
    response: Response = requests.request("DELETE", url, data="", headers=headers)
    return unwrap_delete(response)


@err_chain(IAMTenantException)
def get_issuer(
        headers: dict[str, str],
        account: str = None,
        tenant_id: str = None,
) -> TenantIssuer:
    url = IAMCORE_URL + "/api/v1/tenants/issuers"

    querystring = {
        "account": account,
        "tenant": tenant_id,
    }

    response: Response = requests.request("GET", url, data="", headers=headers, params=querystring)
    return IamEntitiesResponse(TenantIssuer, **unwrap_get(response)).data.pop()


@err_chain(IAMTenantException)
def search_tenant(
        headers: dict[str, str],
        irn: str = None,
        tenant_id: str = None,
        name: str = None,
        display_name: str = None,
        issuer_type: str = None,
        page: int = None,
        page_size: int = None,
        sort: str = None,
        sort_order: SortOrder = None
) -> IamEntitiesResponse[Tenant]:
    url = IAMCORE_URL + "/api/v1/tenants"

    querystring = {
        "irn": str(irn) if irn else None,
        "name": name,
        "displayName": display_name,
        "tenantID": tenant_id,
        "issuerType": issuer_type,
        "page": page,
        "pageSize": page_size,
        "sort": sort,
        "sortOrder": sort_order
    }

    querystring = {
        k: v
        for k, v in querystring.items()
        if v
    }

    response: Response = requests.request("GET", url, data="", headers=headers, params=querystring)
    return IamEntitiesResponse(Tenant, **unwrap_get(response))


@err_chain(IAMException)
def search_all_tenants(auth_headers: dict[str, str], *vargs, **kwargs) -> Generator[Tenant, None, None]:
    return generic_search_all(auth_headers, search_tenant, *vargs, **kwargs)



import unittest

from iamcore.client.application import search_all_applications
from iamcore.client.application_api_key import get_application_api_keys, get_all_applications_api_keys
from iamcore.client.auth import get_token_with_password
from iamcore.client.exceptions import IAMUnauthorizedException, IAMException
from iamcore.client.conf import SYSTEM_BACKEND_CLIENT_ID
from tests.conf import IAMCORE_ROOT_USER, IAMCORE_ROOT_PASSWORD


class GetTokenTestCase(unittest.TestCase):
    def test_token_with_password_ok(self) -> None:
        credentials = get_token_with_password("root", SYSTEM_BACKEND_CLIENT_ID,
                                              IAMCORE_ROOT_USER, IAMCORE_ROOT_PASSWORD)
        self.assertTrue(bool(credentials.access_token))
        self.assertTrue(bool(credentials.refresh_token))
        self.assertTrue(bool(credentials.scope))
        self.assertTrue(bool(credentials.token_type))
        self.assertTrue(bool(credentials.session_state))

    def test_token_with_password_fail(self) -> None:
        with self.assertRaises(IAMUnauthorizedException) as context:
            get_token_with_password("root", SYSTEM_BACKEND_CLIENT_ID, IAMCORE_ROOT_USER, 'nopassword')
        self.assertTrue('Unauthorized:' in context.exception.msg)

    def test_get_api_key(self) -> None:
        application_name = 'kaa'

        token = get_token_with_password("root", SYSTEM_BACKEND_CLIENT_ID, IAMCORE_ROOT_USER, IAMCORE_ROOT_PASSWORD)
        applications = [
            a for a in search_all_applications(token.access_headers, name=application_name)
            if application_name == a.name
        ]
        self.assertGreaterEqual(len(applications), 1)
        for a in applications:
            api_key = next(get_all_applications_api_keys(token.access_headers, a.irn))
            self.assertIsNotNone(api_key)
            self.assertIsNotNone(api_key.api_key)
            self.assertIsInstance(api_key.api_key, str)
            self.assertTrue(api_key.api_key)



import http
from typing import Any

from requests import Response

from http.client import OK, NO_CONTENT, UNAUTHORIZED, BAD_REQUEST, FORBIDDEN, CREATED, CONFLICT


class IAMException(Exception):
    msg: str

    def __init__(self, msg: str):
        self.msg = msg


class IAMUnauthorizedException(IAMException):
    pass


class IAMForbiddenException(IAMException):
    pass


class IAMBedRequestException(IAMException):
    pass


class IAMConflictException(IAMException):
    pass


class IAMTenantException(IAMException):
    pass


class IAMTenantExistsException(IAMTenantException):
    pass


class IAMPolicyException(IAMException):
    pass


class IAMPolicyExistsException(IAMPolicyException):
    pass


class IAMUserException(IAMException):
    pass


class IAMUserExistsException(IAMUserException):
    pass


class IAMGroupException(IAMException):
    pass


class IAMGroupExistsException(IAMGroupException):
    pass


class IAMResourceException(IAMException):
    pass


class IAMResourceExistsException(IAMResourceException):
    pass



class IAMTenantException(IAMException):
    pass


class IAMPolicyException(IAMException):
    pass


class IAMUserException(IAMException):
    pass


class IAMGroupException(IAMException):
    pass


class IAMResourceException(IAMException):
    pass


GET_MAPPING = {
    OK: None,
    UNAUTHORIZED: IAMUnauthorizedException,
    FORBIDDEN: IAMForbiddenException
}

POST_MAPPING = {
    CREATED: None,
    BAD_REQUEST: IAMBedRequestException,
    CONFLICT: IAMConflictException,
    UNAUTHORIZED: IAMUnauthorizedException,
    FORBIDDEN: IAMForbiddenException
}

PUT_MAPPING = {
    NO_CONTENT: None,
    BAD_REQUEST: IAMBedRequestException,
    UNAUTHORIZED: IAMUnauthorizedException,
    FORBIDDEN: IAMForbiddenException
}

DELETE_MAPPING = {
    NO_CONTENT: None,
    UNAUTHORIZED: IAMUnauthorizedException,
    FORBIDDEN: IAMForbiddenException
}

EVALUATE_MAPPING = {
    OK: None,
    BAD_REQUEST: IAMBedRequestException,
    UNAUTHORIZED: IAMUnauthorizedException,
    FORBIDDEN: IAMForbiddenException
}


def unwrap_return_empty(resp: Response, mapping: dict[int, Any]) -> None:
    if resp.status_code not in mapping:
        raise IAMTenantException(f"Unexpected error code: {resp.status_code}")
    mapped_exception = mapping.get(resp.status_code)
    if mapped_exception is None:
        return None
    raise mapped_exception(resp.json()['message'])


def unwrap_return_json(resp: Response, mapping: dict[int, Any]) -> dict[int, Any]:
    if resp.status_code not in mapping:
        raise IAMTenantException(f"Unexpected error code: {resp.status_code}")
    mapped_exception = mapping.get(resp.status_code)
    if mapped_exception is None:
        return resp.json()
    raise mapped_exception(resp.json()['message'])


def unwrap_get(resp: Response, mapping: dict[int, Any] = None) -> dict[int, Any]:
    if mapping is None:
        mapping = GET_MAPPING
    return unwrap_return_json(resp, mapping)


def unwrap_post(resp: Response, mapping: dict[int, Any] = None) -> dict[int, Any]:
    if mapping is None:
        mapping = POST_MAPPING
    return unwrap_return_json(resp, mapping)


def unwrap_put(resp: Response, mapping: dict[int, Any] = None) -> None:
    if mapping is None:
        mapping = PUT_MAPPING
    return unwrap_return_empty(resp, mapping)


def unwrap_patch(resp: Response, mapping: dict[int, Any] = None) -> None:
    if mapping is None:
        mapping = PUT_MAPPING
    return unwrap_return_empty(resp, mapping)


def unwrap_delete(resp: Response, mapping: dict[int, Any] = None) -> None:
    if mapping is None:
        mapping = DELETE_MAPPING
    return unwrap_return_empty(resp, mapping)


def err_chain(error=IAMException):
    def decorator(func):
        def new_func(*args, **kwargs):
            try:
                return func(*args, **kwargs)
            except IAMException as e:
                raise e
            except Exception as e:
                raise error(str(e))

        return new_func

    return decorator



from typing import Generator, Union

import requests
from iamcore.irn import IRN
from requests import Response

from iamcore.client.common import to_snake_case, to_dict, generic_search_all, IamEntitiesResponse
from iamcore.client.conf import IAMCORE_URL
from iamcore.client.exceptions import err_chain, IAMException, unwrap_get


class ApplicationApiKey:
    api_key: str
    state: str
    last_used: str
    created: str
    updated: str

    def __init__(self, **kwargs):
        for k, v in kwargs.items():
            attr = to_snake_case(k)
            setattr(self, attr, v)

    @staticmethod
    def of(item):
        if isinstance(item, ApplicationApiKey):
            return item
        elif isinstance(item, list):
            return [ApplicationApiKey.of(i) for i in item]
        elif isinstance(item, dict):
            return ApplicationApiKey(**item)
        raise IAMException(f"Unexpected response format")

    def to_dict(self):
        return to_dict(self)


@err_chain(IAMException)
def get_application_api_keys(headers: dict[str, str], irn: Union[str, IRN],
                             page: int = 1) -> IamEntitiesResponse[ApplicationApiKey]:
    if isinstance(irn, IRN):
        irn = irn.to_base64()
    if isinstance(irn, str):
        irn = IRN.of(irn).to_base64()

    url = f"{IAMCORE_URL}/api/v1/applications/{irn}/api-keys?page={page}"
    response: Response = requests.request("GET", url, data="", headers=headers)
    return IamEntitiesResponse(ApplicationApiKey, **unwrap_get(response))


@err_chain(IAMException)
def get_all_applications_api_keys(auth_headers: dict[str, str], *vargs, **kwargs) -> Generator[ApplicationApiKey, None, None]:
    return generic_search_all(auth_headers, get_application_api_keys, *vargs, **kwargs)



import logging
from typing import List, Generator

import requests
from iamcore.irn import IRN
from requests import Response

from iamcore.client.conf import IAMCORE_URL
from .common import SortOrder, to_dict, generic_search_all, IamEntitiesResponse, IamEntityResponse
from .exceptions import IAMPolicyException, IAMUnauthorizedException, err_chain, unwrap_post, \
    unwrap_delete, unwrap_get, unwrap_put, IAMException


logger = logging.getLogger(__name__)


class PolicyStatement(object):
    effect: str
    description: str
    resources: List[IRN]
    actions: List[str]

    @staticmethod
    def of(item):
        if isinstance(item, PolicyStatement):
            return item
        elif isinstance(item, dict):
            return PolicyStatement(**item)
        raise IAMPolicyException(f"Unexpected response format")

    def __init__(self, effect: str = None, description: str = None,
                 resources: List[str] = None, actions: List[str] = None):
        self.effect = effect
        self.description = description
        self.resources = [IRN.of(irn) for irn in resources]
        self.actions = actions


class Policy(object):
    id: str
    irn: IRN
    name: str
    description: str
    type: str
    origin: str
    version: str
    statements: List[PolicyStatement]

    def __init__(self, irn: str, statements: List[dict] = None, **kwargs):
        self.irn = IRN.of(irn)
        if isinstance(statements, List):
            self.statements = [
                PolicyStatement.of(item)
                for item in statements
            ]
        for k, v in kwargs.items():
            setattr(self, k, v)

    @staticmethod
    def of(item):
        if isinstance(item, Policy):
            return item
        elif isinstance(item, dict):
            return Policy(**item)
        raise IAMPolicyException(f"Unexpected response format")

    @err_chain(IAMPolicyException)
    def update(self, auth_headers: dict[str, str]) -> None:
        if not auth_headers:
            raise IAMUnauthorizedException(f"Missing authorization headers")
        if not self.id:
            raise IAMPolicyException(f"Missing resource_id or display_name")

        url = IAMCORE_URL + "/api/v1/policies/" + self.id
        payload = to_dict(self)
        headers = {
            "Content-Type": "application/json",
            **auth_headers
        }
        response: Response = requests.request("PUT", url, json=payload, headers=headers)
        unwrap_put(response)

    def delete(self, auth_headers: dict[str, str]) -> None:
        delete_policy(auth_headers, policy_id=self.id)

    def to_dict(self):
        return to_dict(self)


class CreatePolicyRequest(object):
    name: str
    level: str
    tenant_id: str
    description: str
    statements: List[PolicyStatement]

    def __init__(self, name: str, level: str, description: str, tenant_id: str = None):
        self.name = name
        self.level = level
        self.tenant_id = tenant_id
        self.description = description
        self.statements = []
        if level == 'tenant' and not tenant_id:
            logger.warning("Missing tenant_id for tenant level policy")


    def with_statement(self, effect: str, description: str, resources: List[str], actions: List[str]):
        self.statements.append(PolicyStatement(
            effect=effect, description=description, resources=resources, actions=actions))
        return self

    @err_chain(IAMPolicyException)
    def create(self, auth_headers: dict[str, str]) -> Policy:
        return create_policy(auth_headers, self)

    def to_dict(self):
        return {
            'name': self.name,
            'level': self.level,
            'description': self.description,
            'tenantID': self.tenant_id,
            'statements': to_dict(self.statements),
        }


@err_chain(IAMPolicyException)
def create_policy(auth_headers: dict[str, str], payload: CreatePolicyRequest) -> Policy:
    url = IAMCORE_URL + "/api/v1/policies"
    payload = payload.to_dict()

    headers = {
        "Content-Type": "application/json",
        **auth_headers
    }
    response: Response = requests.request("POST", url, json=payload, headers=headers)
    return IamEntityResponse(Policy, **unwrap_post(response)).data


@err_chain(IAMPolicyException)
def delete_policy(auth_headers: dict[str, str], policy_id: str) -> None:
    if not auth_headers:
        raise IAMUnauthorizedException(f"Missing authorization headers")
    if not policy_id:
        raise IAMPolicyException(f"Missing resource_id")

    url = IAMCORE_URL + "/api/v1/policies/" + IRN.of(policy_id).to_base64()
    headers = {
        "Content-Type": "application/json",
        **auth_headers
    }
    response: Response = requests.request("DELETE", url, data="", headers=headers)
    unwrap_delete(response)


@err_chain(IAMPolicyException)
def search_policy(
        headers: dict[str, str],
        irn: str = None,
        name: str = None,
        description: str = None,
        page: int = None,
        page_size: int = None,
        sort: str = None,
        sort_order: SortOrder = None
) -> IamEntitiesResponse[Policy]:
    url = IAMCORE_URL + "/api/v1/policies"

    querystring = {
        "irn": irn,
        "name": name,
        "description": description,
        "page": page,
        "pageSize": page_size,
        "sort": sort,
        "sortOrder": sort_order
    }

    querystring = {
        k: v
        for k, v in querystring.items()
        if v
    }

    response = requests.request("GET", url, data="", headers=headers, params=querystring)
    return IamEntitiesResponse(Policy, **unwrap_get(response))


@err_chain(IAMException)
def search_all_policies(auth_headers: dict[str, str], *vargs, **kwargs) -> Generator[Policy, None, None]:
    return generic_search_all(auth_headers, search_policy, *vargs, **kwargs)



import logging
from typing import List, Generator, Union, Any

import requests
from iamcore.irn import IRN
from requests import Response

from iamcore.client.common import IamEntitiesResponse, generic_search_all
from iamcore.client.conf import IAMCORE_URL
from iamcore.client.exceptions import EVALUATE_MAPPING, unwrap_return_empty, IAMException, unwrap_return_json

logger = logging.getLogger(__name__)


def authorize(
        authorization_headers: dict[str, str],
        principal_irn: IRN,
        account_id: str,
        application: str,
        tenant_id: str,
        resource_type: str,
        resource_path: str,
        action: str,
        resource_ids: Union[List[str], None] = None) -> List[IRN]:
    if not action:
        raise IAMException("Action must be defined")
    tenant_id = tenant_id or principal_irn.tenant_id
    account_id = account_id or principal_irn.account_id
    if resource_ids:
        resources_irn_list = [
            IRN.create(
                account_id=account_id,
                application=application,
                tenant_id=tenant_id,
                resource_type=resource_type,
                resource_path=resource_path,
                resource_id=resource_id)
            for resource_id in resource_ids
        ]
        logger.debug(f"Going to evaluate {authorization_headers} {action} {resources_irn_list}")
        evaluate(authorization_headers, action, resources_irn_list)
        return resources_irn_list
    logger.debug("Going to evaluate by type")
    return list(evaluate_all_resources(authorization_headers, application, action, resource_type))


def evaluate(auth_headers: dict[str, str], action: str, resources: List[IRN]) -> None:
    url = IAMCORE_URL + "/api/v1/evaluate"
    payload = {
        "action": action,
        "resources": [str(r) for r in resources if r]
    }
    headers = {
        "Content-Type": "application/json",
        **auth_headers
    }
    logger.debug(f"Going to evaluate resources: POST {url}, json={payload}")
    response: Response = requests.request("POST", url, json=payload, headers=headers)
    return unwrap_return_empty(response, EVALUATE_MAPPING)


def evaluate_resources(
        auth_headers: dict[str, str],
        application: str,
        action: str,
        resource_type: str,
        page: int = 1,
        page_size: int = 100) -> IamEntitiesResponse[IRN]:
    url = IAMCORE_URL + "/api/v1/evaluate/resources"
    payload = {
        "application": application,
        "action": action,
        "resourceType": resource_type
    }
    querystring = {
        "page": page,
        "pageSize": page_size,
    }
    headers = {
        "Content-Type": "application/json",
        **auth_headers
    }
    logger.debug(f"Going to evaluate resource type: POST {url}, json={payload}, params={querystring}")
    response: Response = requests.request("POST", url, json=payload, headers=headers, params=querystring)
    return IamEntitiesResponse(IRN, **unwrap_return_json(response, EVALUATE_MAPPING))


def evaluate_all_resources(auth_headers: dict[str, str], *vargs, **kwargs) -> Generator[IRN, None, None]:
    return generic_search_all(auth_headers, evaluate_resources, *vargs, **kwargs)



import http.client
from uuid import UUID

import requests

from .exceptions import IAMException, IAMUnauthorizedException
from iamcore.client.conf import IAMCORE_ISSUER_URL


class TokenResponse:
    access_token: str
    expires_in: int
    refresh_expires_in: int
    refresh_token: int
    token_type: str
    not_before_policy: int
    session_state: UUID
    scope: str

    def __init__(self, **kwargs):
        for k, v in kwargs.items():
            if '-' in k:
                setattr(self, k.replace("-", "_"), v)
            else:
                setattr(self, k, v)

    @property
    def access_headers(self) -> dict[str, str]:
        return {
            "Authorization": "Bearer " + self.access_token
        }


def get_api_key_auth_headers(api_key: str):
    return {
        'X-iamcore-API-Key': api_key
    }


def get_token_with_password(realm: str, client_id, username: str, password: str) -> TokenResponse:
    url = f"{IAMCORE_ISSUER_URL.strip()}/realms/{realm}/protocol/openid-connect/token"
    payload = f"grant_type=password&client_id={client_id}&username={username}&password={password}"
    headers = {
        "Content-Type": "application/x-www-form-urlencoded"
    }
    try:
        response = requests.request("POST", url, data=payload, headers=headers)
        if response.status_code == http.client.OK:
            return TokenResponse(**response.json())
        elif response.status_code == http.client.UNAUTHORIZED:
            raise IAMUnauthorizedException(f"Unauthorized: {response.json()}")
        raise IAMUnauthorizedException(f"Unexpected error code: {response.status_code}")
    except IAMException as e:
        raise e
    except Exception as e:
        raise IAMException(f"Failed to get auth token with exception: {e}")



import os
from dotenv import load_dotenv

load_dotenv()

IAMCORE_URL: str = os.getenv("IAMCORE_URL", "http://iam.kaatech.com")
IAMCORE_ISSUER_URL: str = os.environ.get('IAMCORE_ISSUER_URL', 'http://iam.kaatech.com/auth')
SYSTEM_BACKEND_CLIENT_ID: str = os.environ.get('SYSTEM_BACKEND_CLIENT_ID')
